from accelforge.frontend.renames import TensorName


import itertools
from enum import Enum

import accelforge.frontend.arch as arch
from accelforge.frontend.arch._flattened_arch import FlattenedArch
from accelforge.util._frozenset import oset
from accelforge.frontend.mapping import (
    MappingNode,
    Toll,
    Temporal,
    Spatial,
    TensorHolder,
    Loop,
)
from accelforge.frontend.workload import (
    Einsum,
    RankVariable,
    Workload,
)

# =================================================================================================
# Insert loops
# =================================================================================================


class LowerChoice(Enum):
    YES = 0
    NO = 1
    OPTIONAL = 2


def insert_temporal_loops(
    mapping: list[TensorHolder],
    einsum: Einsum,
    first_memory: arch.Memory,
    rank_variable_bounds: dict[RankVariable, int],
    ranks_with_tile_pattern: set,
    workload: Workload,
    _can_lower_outermost_memory: bool,
    flattened_arch: FlattenedArch,
    max_fused_loops: int,
    fanouts: dict[str, int],
    fusable_tensors: set[TensorName],
    intermediate_tensors: set[TensorName],
    let_non_intermediate_tensors_respawn_in_backing_storage: bool,
    explore_loop_orders: bool,
):
    # First establish insertion points. Insertion points are:
    # - Below the last instance of the first memory
    # - Between any two TensorHolder nodes
    # - After the last TensorHolder node

    # The following logic is really just to make sure that all the storage nodes for the
    # outermost memory are together at the beginning of the split mapping. After that,
    # each entries in the split mapping has a single TensorHolder.
    split_mapping: list[list[TensorHolder]] = []
    inserted = oset()  # for tracking which element of `mapping` is handled

    # The first group is for persistent tensors
    storage_nodes_above_any_loop = []
    for m in mapping:
        if id(m) in inserted:
            continue
        if m.persistent:
            storage_nodes_above_any_loop.append(m)
            inserted.add(id(m))

    # TODO: we have a misnomer: what we call `first_memory` in code is the one that
    # `_can_lower_outermost_memory` refers to.
    if not _can_lower_outermost_memory:
        for m in mapping:
            if id(m) in inserted:
                continue
            if m.component == first_memory.name:
                storage_nodes_above_any_loop.append(m)
                inserted.add(id(m))
    split_mapping.append(storage_nodes_above_any_loop)

    # All other storage nodes may have loops between them
    for m in mapping:
        if id(m) in inserted:
            continue
        split_mapping.append([m])

    for s in split_mapping:
        # Within each split mapping group, sort by arch levels.
        # This can help create places to put spatial loops
        s.sort(key=lambda tensor_holder: flattened_arch.index(tensor_holder.component))

    if sum(map(len, split_mapping)) != len(mapping):
        raise RuntimeError("BUG: number of storage nodes post-split != original")

    for i in range(len(split_mapping) - 1):
        for j in range(i + 1, len(split_mapping)):
            for m_i in split_mapping[i]:
                for m_j in split_mapping[j]:
                    if mapping.index(m_i) > mapping.index(m_j):
                        raise RuntimeError(
                            f"BUG: node {m_i.compact_str()} and "
                            f"{m_j.compact_str()} inverted post-splitting\n"
                            f"Original mapping: {[m.compact_str() for m in mapping]}\n"
                            f"Split mapping: {[[m.compact_str() for m in s] for s in split_mapping]}\n"
                        )

    # These Einsum properties are recalculated since Einsum is mutable
    # We're pre-computing and reusing for efficiency
    tensor2fully_relevant_rank_vars = einsum.tensor2directly_indexing_rank_variables
    tensor2partially_relevant_rank_vars = (
        einsum.tensor2expression_indexing_rank_variables
    )
    for k, v in tensor2partially_relevant_rank_vars.items():
        tensor2partially_relevant_rank_vars[k] = v - tensor2fully_relevant_rank_vars[k]
    tensor2irrelevant_rank_vars = einsum.tensor2irrelevant_rank_variables
    tensors = einsum.tensor_names

    is_fused_loops = True
    seen_tensors = oset()
    choices = []
    lowering_choices: list[tuple[bool, ...]] = []

    def _get_next_storages(i: int, toll_allowed: bool = False) -> list[TensorHolder]:
        for j in range(i + 1, len(split_mapping)):
            assert (
                len(split_mapping[j]) <= 1
            ), f"Mapping: {[m.compact_str() for m in mapping]}"
            # We don't add loops before tolls since they don't reuse things
            if isinstance(split_mapping[j][0], Toll) and not toll_allowed:
                continue
            return split_mapping[j]
        return []

    for i, prev_storages in enumerate(split_mapping):
        # =============================================================================
        # Choose what temporal loops to insert between prev_storages and the next
        # TensorHolder node(s).
        # =============================================================================

        next_storages = _get_next_storages(i)
        next_anything = _get_next_storages(i, toll_allowed=True)

        for s in prev_storages:
            # No tensor holders must mix backing/non-backing tensors.
            assert not s._backing or all(t in s._backing for t in s.tensors)
            # One tensor per holder
            assert len(s.tensors) == 1

        rank_variables = einsum.rank_variables
        # rank_variables = {r for r in rank_variables if rank_variable_bounds[r] > 1}
        seen_tensors |= oset.union(*(oset(t.tensors) for t in prev_storages), oset())
        is_fused_loops = is_fused_loops and len(fusable_tensors - seen_tensors) > 0
        prev_tensors = oset.union(oset(), *(oset(t.tensors) for t in prev_storages))
        next_persistent = oset.union(
            oset(), *(oset(t.tensors) for t in next_storages if t.persistent)
        )

        max_fanout_before = max(
            [fanouts[s2.component] for s in split_mapping[:i] for s2 in s],
            default=float("inf"),
        )
        cur_fanout = oset(fanouts[s2.component] for s2 in prev_storages)
        next_fanout = oset(fanouts[s2.component] for s2 in next_anything)
        if len(cur_fanout) == 0:  # Happens if we're inserting above all storage nodes
            cur_fanout.add(1)
        if len(next_fanout) == 0:  # Happens if we're inserting below all storage nodes
            next_fanout.add(float("inf"))

        # These are used to check if fanouts increased and therefore spatial loops
        # may be placed, so we use the maximum.
        assert len(cur_fanout) > 0
        assert len(next_fanout) > 0
        cur_fanout = max(cur_fanout)
        next_fanout = max(next_fanout)

        # Can't have loops above persistent tensor holders
        if next_persistent:
            rank_variables &= oset()

        # No recomputation: If we haven't seen a tensor yet, must only iterate over
        # fully-relevant rank variables.
        check_tensors = tensors
        if let_non_intermediate_tensors_respawn_in_backing_storage:
            check_tensors = intermediate_tensors

        for t in check_tensors - seen_tensors:
            rank_variables &= tensor2fully_relevant_rank_vars[t]

        if max_fused_loops == 0 and (fusable_tensors - seen_tensors):
            rank_variables &= oset()

        # The fanout for a prior node may be placed here, so spatial nodes may be moved
        # here
        someone_elses_spatials_may_be_placed_below = (
            next_fanout < cur_fanout and max_fanout_before < cur_fanout
        )

        # If the fanout is about to increase, then spatial loops may be placed below the
        # current node. There may have been constrained temporal loops earlier that need
        # to be placed here, so we won't prohibit any loops. TODO:
        # CONTIGUOUS_ITERATION_SPACE_DISCUSSION: This causes all loops to be added, but
        # really we only need to re-add the ones that may conflict with a spatial loop.
        if someone_elses_spatials_may_be_placed_below:
            pass
        else:
            # Optimality-preserving optimization: Loops below tolls aren't helpful
            # because there is no storage. Ctrl-F for
            # CONTIGUOUS_ITERATION_SPACE_DISCUSSION: Can't do this if we may put another
            # node's spatial loops below this one, because lowering would add move the
            # spatials down, which would constrain the temporals due to spatial-temporal
            # crossing.
            if prev_storages and isinstance(prev_storages[0], Toll):
                rank_variables &= oset()

            # Optimality-preserving optimization: We can trivially lower non-backing
            # TensorHolder nodes through fully-relevant loops. Can't do this if the
            # loops are fused because that'd add loops to the compatibility.
            # CONTIGUOUS_ITERATION_SPACE_DISCUSSION: Can't do this if we may put another
            # node's spatial loops below this one, because lowering would add move the
            # spatials down, which would constrain the temporals due to spatial-temporal
            # crossing.
            for s in prev_storages:
                for t in s.tensors:
                    if t not in s._backing and not s._must_be_here:
                        rank_variables -= tensor2fully_relevant_rank_vars[t]

            # Optimality-preserving optimization: We can trivially raise TensorHolder
            # nodes through irrelevant unfused loops. Can't do this if the loops are
            # fused because that'd increase the lifetime of the TensorHolder node. Can't
            # do this if the irrelevant rank variables are partially-relevant to the
            # previous tensors, since that affects the permutation. See
            # CONTIGUOUS_ITERATION_SPACE_DISCUSSION: Can't do this if we may put another
            # node's spatial loops above this one, because raising would add move the
            # temporals down, which would constrain them due to spatial-temporal
            # crossing.
            if not is_fused_loops:
                for s in next_storages:
                    if not s._must_be_here:
                        for t in s.tensors:
                            rvs = tensor2irrelevant_rank_vars[t]
                            for t2 in prev_tensors:
                                rvs -= tensor2partially_relevant_rank_vars[t2]
                            rank_variables -= rvs

        # =============================================================================
        # Determine whether to lower TensorHolder nodes through partially-relevant
        # loops.
        # =============================================================================
        partially_relevant_to_previous = rank_variables & oset.union(
            oset(), *(tensor2partially_relevant_rank_vars[t] for t in prev_tensors)
        )
        permutable_partially_relevant = oset()

        # NOTE: If the lowering logic for backing TensorHolders is updated & we can
        # lower through >1 loops, then also update label_fused_loops
        for s in prev_storages:
            partially_relevant_to_previous = oset.union(
                oset(), *(tensor2partially_relevant_rank_vars[t] for t in s.tensors)
            )
            partially_relevant_to_previous &= rank_variables
            lowerable_backing = (
                _can_lower_outermost_memory or s.component != first_memory.name
            )

            # Persistent. Must be at the top of the mapping.
            if s.persistent:
                lowering_choices.append((False,))
            # Don't lower our own reservations through someone else's spatial loops.
            elif someone_elses_spatials_may_be_placed_below:
                lowering_choices.append((False,))
            # Processing stage. Lowering doesn't matter. Don't lower.
            elif isinstance(s, Toll):
                lowering_choices.append((False,))
            # Previous is backing and there's partially-relevant rank variables. May
            # want to lower to reduce memory footprint, or raise to reduce number of
            # fused loops.
            elif s._backing and lowerable_backing and partially_relevant_to_previous:
                lowering_choices.append((False, True))
                permutable_partially_relevant |= partially_relevant_to_previous
            # No backing in previous. No cost to lowering. Lower all
            elif not s._backing:
                # print(f'Can lower {s.tensors} in {s.component}')
                lowering_choices.append((True,))
                permutable_partially_relevant |= partially_relevant_to_previous
            # Previous TensorHolder is backing but not lowerable or there are no
            # partially relevant rank vars.
            else:
                lowering_choices.append((False,))

        # =============================================================================
        # Create loop order and lowering choices
        # =============================================================================

        can_lower = any(any(c) for c in lowering_choices)

        # Create canonical loop orders that avoids repeating reuse patterns.
        choices.append(
            list(
                canonical_loop_orders(
                    rank_variables,
                    permutable_partially_relevant,
                    can_lower,
                    explore_loop_orders,
                )
            )
        )

    # ==================================================================================
    # Iterate over all possible mappings
    # ==================================================================================

    # TODO: Optimization: If we can optionally lower a tensor & the loop below it is
    # not something through which we can lower for a given permutation, skip options
    # that lower that tensor because they get the same result as not lowering the
    # tensor.
    n_loop_orders = len(list(itertools.product(*choices)))
    for loop_orders in itertools.product(*choices):
        full_mapping = []
        for prev_storages, loop_order in zip(split_mapping, loop_orders):
            full_mapping.extend(prev_storages)
            full_mapping.extend(Temporal(rank_variable=r) for r in loop_order)

        storages = [node for node in full_mapping if isinstance(node, TensorHolder)]
        assert len(lowering_choices) == len(storages)
        for lowering_choice in itertools.product(*lowering_choices):
            for lower, node in zip(lowering_choice, storages):
                node._lower = lower

            yield list(full_mapping), n_loop_orders


def insert_spatial_loops(
    mapping: list[MappingNode],
    einsum: Einsum,
    flattened_arch: FlattenedArch,
    intermediate_tensors: set[TensorName],
):
    nodes_with_fanout = [n for n in flattened_arch if n.get_fanout() > 1]
    tensor2fully_relevant_rank_vars = einsum.tensor2directly_indexing_rank_variables
    simple_rank_variables = einsum._simple_rank_variables

    for node in nodes_with_fanout:
        # Insert spatials below the lowest storage node whose component is
        # above the fanout in the arch, and below any temporal loops in the
        # same block.
        insertion_point = _idx_below_lowest_tensor_holder_with_component_above_fanout(
            node, mapping, flattened_arch
        )
        while insertion_point < len(mapping) and isinstance(
            mapping[insertion_point], Temporal
        ):
            insertion_point += 1

        # No recomputation: If we haven't seen a tensor yet, must only iterate over
        # fully-relevant rank variables.
        rank_variables = einsum.rank_variables
        for t in intermediate_tensors - _tensors_seen_above_point(
            insertion_point, mapping
        ):
            rank_variables &= tensor2fully_relevant_rank_vars[t]

        for fanout_dim in node.spatial:
            for r in sorted(rank_variables):
                s = Spatial(
                    rank_variable=r,
                    name=fanout_dim.name,
                    component_object=node,
                    component=node.name,
                )
                if insertion_point == len(mapping):
                    mapping.append(s)
                else:
                    mapping.insert(insertion_point, s)


def _tensors_seen_above_point(idx, mapping):
    seen_tensors = oset()
    for i in range(idx):
        node = mapping[i]
        if not isinstance(node, TensorHolder):
            continue
        seen_tensors |= oset(node.tensors)
    return seen_tensors


def _idx_below_lowest_tensor_holder_with_component_above_fanout(
    fanout_node, mapping, flattened_arch: FlattenedArch
):
    """Return the index right after the lowest TensorHolder whose component
    is above the fanout in the arch. If none found, returns len(mapping)."""
    fanout_arch_idx = flattened_arch.index(fanout_node.name)
    result = 0
    for i in range(len(mapping)):
        if not isinstance(mapping[i], TensorHolder):
            continue
        if flattened_arch.index(mapping[i].component) < fanout_arch_idx:
            result = i + 1
    return result


def label_imperfect_tile_shapes(mapping, einsum: Einsum, job):
    allow_all = job.allow_imperfect_all
    imperfect_spatial = allow_all or job.explore_imperfect_spatial_loops
    imperfect_temporal = allow_all or job.explore_imperfect_temporal_loops
    simple_ranks = einsum._simple_rank_variables

    rv2nodes = {}
    for node in mapping:
        if isinstance(node, Loop) and node.rank_variable in simple_ranks:
            rv2nodes.setdefault(node.rank_variable, []).append(node)

    for _, nodes in rv2nodes.items():
        for i, node in enumerate(nodes):
            # (A) We can get full expressiveness by unlocking options for the PREVIOUS
            # loop, (B) which is also much faster to search. However, if we're the
            # outermost loop for this rank variable, then unlock options for this loop.

            # (A) reasoning: Loop bound = ceil(outer size / this size). Imperfect can
            # relax options, so we choose outer size.
            #
            # (B) reasoning: For spatial loops, we know the fanout once we pick outer
            # size AND inner size. We enumerate inner to outer, so this effectively
            # defers imperfect factorization until we know the exact fanout, and can
            # immediately use it to prune too-large fanouts.
            target = max(i - 1, 0)
            if isinstance(node, Spatial) and imperfect_spatial:
                nodes[target]._may_cause_imperfect = True
            if isinstance(node, Temporal) and imperfect_temporal:
                nodes[target]._may_cause_imperfect = True


def canonical_loop_orders(
    rank_variables: set[RankVariable],
    partially_relevant_to_previous: set[RankVariable],
    can_lower: bool,
    explore_loop_orders: bool,
):
    """Generate loop orders that result in unique reuse patterns."""
    # Only the first partially-relevant rank variable matters is a meaningful
    # choice because lowering only happens through at most one rank var.
    if not partially_relevant_to_previous or not can_lower:
        yield tuple(sorted(rank_variables))
        return

    for first_rank_var in partially_relevant_to_previous:
        rest_of_partially_relevant = partially_relevant_to_previous - {first_rank_var}
        rest_rank_vars = rank_variables - partially_relevant_to_previous
        # Since order does not matter, we choose alphabetical order as canonical.
        yield (
            (first_rank_var,)
            + tuple(sorted(rest_of_partially_relevant))
            + tuple(sorted(rest_rank_vars))
        )
        if not explore_loop_orders:
            return



import copy
from collections.abc import Generator
from itertools import chain, combinations
import logging

import accelforge.frontend.arch as arch
from accelforge.frontend.mapping import Storage, TensorHolder, Toll
from accelforge.frontend.workload import TensorName, SymbolTable

from accelforge.util._basetypes import EvalableList
from accelforge.util.exceptions import EvaluationError
from accelforge.util._setexpressions import InvertibleSet
from accelforge.util._frozenset import oset


def make_tensor_choices_one_level(
    node: arch.Leaf,
    symbol_table: dict[str, InvertibleSet],
    persistent_tensors: set[TensorName],
    seen_tensors: set[TensorName] = (),
    is_copy_op: bool = False,
    einsum_name: str = None,
    prioritize_reuse_of_unfused_tensors: bool = False,
) -> Generator[tuple[list[TensorHolder], SymbolTable, set[TensorName]], None, None]:
    """
    Generate combinations of TensorHolder nodes based on keep and bypass
    constraints.

    Each generated list contains TensorHolder nodes for single tensors.
    """
    assert "All" in symbol_table
    tensors = symbol_table["All"]

    if not isinstance(node, arch.TensorHolder):
        yield [], symbol_table, oset(seen_tensors)
        return

    if isinstance(node, arch.Memory):
        target_type = Storage
    elif isinstance(node, arch.Toll):
        target_type = Toll
    elif isinstance(node, arch.Dummy):
        yield [], symbol_table, oset(seen_tensors)
        return
    else:
        raise ValueError(f"Unexpected tensor holder type: {type(node)}")

    new_symbol_table = copy.copy(symbol_table)

    node: arch.TensorHolder = copy.copy(node)
    if not node.tensors.tensor_order_options and prioritize_reuse_of_unfused_tensors:
        node.tensors = copy.copy(node.tensors)
        node.tensors.tensor_order_options.append(EvalableList(["Above", "~Above"]))
    try:
        node.tensors: arch.Tensors = node.tensors._eval_expressions(
            symbol_table=symbol_table,
            musteval_tryeval_to=True,
            must_copy=False,
            location=f"arch.{node.name}.tensors",
        )[0]
    except EvaluationError as e:
        e.add_field(f"Einsum {einsum_name} arch.{node.name}.tensors")
        raise e

    if not node.enabled:
        yield [], symbol_table, oset(seen_tensors)
        return

    must_keep = tensors.to_my_space(node.tensors.keep | node.tensors.back)
    may_keep = tensors.to_my_space(node.tensors.may_keep)
    may_keep -= must_keep

    # Tolls are passthrough and cannot be the outermost level storing a tensor.
    if isinstance(node, arch.Toll):
        above = tensors.to_my_space(symbol_table.get("Above", oset()))
        must_keep &= above
        may_keep &= above

    if seen_tensors & oset(node.tensors.back):
        return

    if must_keep - tensors:
        raise KeyError(
            f"Keep constraint for {node.name} includes tensors that are "
            f"not in the workload: {must_keep - new_symbol_table['All']}"
        )
    if may_keep - tensors:
        raise KeyError(
            f"Bypass constraint for {node.name} includes tensors that are "
            f"not in the workload: {may_keep - tensors.full_space}"
        )

    logging.info(
        f"\t\t{node.name} must keep {sorted(must_keep)}, may keep {sorted(may_keep)}"
    )

    # No reuse in copy operations, so no need to keep tensors in more places
    if is_copy_op:
        may_keep -= tensors.to_my_space(seen_tensors)

    for subset in powerset(sorted(may_keep, key=str)):
        # Make keep choice & update symbol table
        subset = tensors.to_my_space(oset(subset))
        keep_choice = tensors.to_my_space(subset | must_keep)
        # Below line is so users can do MainMemory().tensors() or MainMemory.tensors
        new_symbol_table[node.name] = keep_choice
        new_symbol_table["Above"] = symbol_table["Above"] | keep_choice
        new_seen_tensors = seen_tensors | oset(keep_choice)

        # Make sure they're all tensors
        assert all(isinstance(k, TensorName) for k in keep_choice)
        keep_choice = keep_choice.to_my_space(oset(copy.copy(t) for t in keep_choice))
        nodes = []

        # Create storage nodes. Sort them to keep this deterministic. Ordering is done
        # later.
        for t in sorted(keep_choice, key=str):
            nodes.append(
                target_type(tensors=[t], component=node.name, component_object=node)
            )
            if t not in seen_tensors:
                nodes[-1]._backing.add(t)
                nodes[-1]._must_keep_tensors = [t]
                nodes[-1].persistent = t in persistent_tensors
            elif t in must_keep:
                nodes[-1]._must_keep_tensors = [t]

        yield nodes, new_symbol_table, new_seen_tensors


def make_storage_choices_all_levels(
    nodes: list[TensorHolder],
    symbol_table: dict[str, InvertibleSet],
    persistent_tensors: set[TensorName],
    seen_tensors: set[TensorName],
    is_copy_op: bool,
    einsum_name: str,
    prioritize_reuse_of_unfused_tensors: bool,
) -> Generator[tuple[dict[str, list[TensorHolder]], SymbolTable], None, None]:
    """
    Generate combinations of TensorHolder nodes based on keep and bypass
    constraints.

    Each generated dict maps memory name to a list of TensorHolder nodes for
    single tensors.
    """
    seen_tensors = oset() if seen_tensors is None else seen_tensors
    if len(nodes) == 0:
        yield dict(), symbol_table
        return

    if "Above" not in symbol_table:
        symbol_table["Above"] = symbol_table["Nothing"]

    for choice, symbol_table, new_seen_tensors in make_tensor_choices_one_level(
        node=nodes[0],
        symbol_table=symbol_table,
        persistent_tensors=persistent_tensors,
        seen_tensors=seen_tensors,
        is_copy_op=is_copy_op,
        einsum_name=einsum_name,
        prioritize_reuse_of_unfused_tensors=prioritize_reuse_of_unfused_tensors,
    ):
        for subchoices, symbol_table in make_storage_choices_all_levels(
            nodes=nodes[1:],
            symbol_table=symbol_table,
            persistent_tensors=persistent_tensors,
            seen_tensors=new_seen_tensors,
            is_copy_op=is_copy_op,
            einsum_name=einsum_name,
            prioritize_reuse_of_unfused_tensors=prioritize_reuse_of_unfused_tensors,
        ):
            yield {**subchoices, nodes[0].name: choice}, symbol_table


def powerset(iterable):
    s = list(iterable)
    return chain.from_iterable(combinations(s, r) for r in range(len(s) + 1))



from typing import Generator, List, Tuple

from accelforge.util._frozenset import oset
from accelforge.frontend.mapping import (
    Compute,
    Mapping,
    MappingNode,
    Pipeline,
    Sequential,
)
from accelforge.frontend.workload import Workload


def get_paths(root: Mapping) -> Generator[Tuple[MappingNode, Compute], None, None]:
    """
    Given a MappingNode, get the paths to all all leaves in post-order.

    :param root:    The root of the child exploration.

    :type root:     MappingNode

    :returns:       A generator of all the MappingNodes to a Compute leaf.
    :rtype:         Generator[List[MappingNode]]
    """
    cur_path: List[MappingNode] = []
    for node in root.nodes:
        cur_path.append(node)
        match node:
            # Pipelines or sequentials should have their paths expanded.
            # Mappings naturally get expanded.
            case Mapping() | Pipeline() | Sequential():
                for child in node.nodes:
                    for subpath in get_paths(child):
                        yield tuple(cur_path) + subpath
            # Computes are leaves so should get a yield here.
            case Compute():
                yield tuple(cur_path)
            # Not implemented so continue.
            case _:
                # TODO: Check this is correct
                continue
                raise NotImplementedError(
                    f"{type(node)} does not have type elucidation.\n"
                    f"---\n"
                    f"node={node}"
                )


def get_leaves(mapping: Mapping, is_path):
    if is_path:
        yield mapping[-1]
        return
    for node in mapping:
        if isinstance(node, Pipeline) or isinstance(node, Sequential):
            for child in node.children:
                yield from get_leaves(child, is_path)
        elif isinstance(node, Compute):
            yield node


def get_intermediate_tensors(workload: Workload):
    result = oset()
    for einsum in workload.einsum_id_to_name():
        written_tensors = workload.einsums[einsum].output_tensor_names
        for tensor in written_tensors:
            reader_einsums = workload.reader_einsums(tensor)
            for reader in reader_einsums:
                if reader in workload.einsum_id_to_name():
                    result.add(tensor)
                    break

    return result



from docutils import nodes
from docutils.parsers.rst import Directive
import importlib
import inspect


class IncludeFunctions(Directive):
    required_arguments = 1  # fully-qualified name

    def run(self):
        fqname = self.arguments[0]
        parts = fqname.split(".")

        # --- progressively import the longest valid module ---
        module = None
        for i in range(len(parts), 0, -1):
            try:
                module = importlib.import_module(".".join(parts[:i]))
                rest = parts[i:]
                break
            except ImportError:
                continue

        if module is None:
            return []

        obj = module
        for part in rest:
            if hasattr(obj, part):
                obj = getattr(obj, part)
            else:
                return []

        # --- Collect all functions ---
        functions = {}  # {func_name: {'signature': ..., 'doc': ...}}

        # Get all members
        for name, member in inspect.getmembers(obj):
            # Skip underscore-prefixed and special functions
            if self._should_skip_function(name):
                continue

            # Check if it's a function or method
            if inspect.isfunction(member) or inspect.ismethod(member):
                doc = self._extract_summary_doc(member)
                functions[name] = {"signature": self._get_signature(member), "doc": doc}

        # --- Build bullet list ---
        if not functions:
            return []

        bullet_list = nodes.bullet_list()
        for func_name in sorted(functions.keys()):
            func_info = functions[func_name]
            list_item = nodes.list_item()
            para = nodes.paragraph()

            # Function name as :py:func: role for clickable links
            from sphinx.addnodes import pending_xref

            refnode = pending_xref(
                "",
                refdomain="py",
                reftype="func",
                reftarget=fqname + "." + func_name,
                refwarn=True,
            )
            # refnode += nodes.literal('', f"{func_name}{func_info['signature']}", classes=['xref', 'py', 'py-func'])
            refnode += nodes.literal(
                "", f"{func_name}()", classes=["xref", "py", "py-func"]
            )
            para += refnode

            # Docstring
            if func_info["doc"]:
                para += nodes.Text(f": {func_info['doc']}")

            list_item += para
            bullet_list += list_item

        return [bullet_list]

    def _should_skip_function(self, func_name):
        """Check if a function should be skipped."""
        return func_name.startswith("_") or func_name.startswith("__")

    def _get_signature(self, func):
        """Get the function signature as a string."""
        try:
            sig = inspect.signature(func)
            return str(sig)
        except (ValueError, TypeError):
            return "()"

    def _extract_summary_doc(self, func):
        """Extract the summary part of the docstring, stopping at section headers."""
        # Get raw docstring
        doc = getattr(func, "__doc__", None)
        if not doc:
            return None

        # Clean and process the docstring
        doc = inspect.cleandoc(doc)

        section_headers = [
            "Parameters",
            "Returns",
            "Postcondition",
            "Raises",
            "Yields",
            "Examples",
            "Notes",
            "See Also",
            "Attributes",
            "Methods",
            "References",
            "Warnings",
            "Args",
            "Return",
            "Keyword Arguments",
            "Other Parameters",
        ]

        lines = doc.split("\n")
        summary_lines = []

        for line in lines:
            stripped = line.strip()

            # Stop at empty line followed by section header pattern
            if not stripped:
                # Check if we've already collected some content
                if summary_lines:
                    # Peek ahead to see if next non-empty line is a section
                    continue

            # Check if this line is a section header
            if stripped in section_headers:
                break

            # Check if this line ends with a colon and might be a section header
            if stripped.endswith(":"):
                header_candidate = stripped[:-1].strip()
                if header_candidate in section_headers:
                    break

            # Check for common docstring section patterns (underlined headers)
            if stripped and all(c in "-=~" for c in stripped):
                break

            summary_lines.append(line)

        # Join lines and clean up
        summary = "\n".join(summary_lines).strip()

        # Replace multiple newlines/spaces with single space for better formatting
        summary = " ".join(summary.split())

        return summary if summary else None


def setup(app):
    app.add_directive("include-functions", IncludeFunctions)



from dataclasses import dataclass
from typing import Any
from functools import reduce, lru_cache
from operator import mul

import sympy

from accelforge.frontend.workload import (
    TensorName,
    Einsum,
    EinsumName,
    Workload,
    Rank,
    RankVariable,
)
from ._isl import get_rank_variable_bounds


@lru_cache(maxsize=4096)
def parse_expr(s: str) -> sympy.Expr:
    return sympy.parsing.sympy_parser.parse_expr(s)


def get_projection_expr(einsum: Einsum, tensor: TensorName) -> dict[Rank, sympy.Expr]:
    projection = einsum.tensor_accesses[tensor].projection
    return {
        rank_name: parse_expr(proj_str) for rank_name, proj_str in projection.items()
    }


class Irrelevant:
    pass


@dataclass
class Relevant:
    rank: Any


@dataclass
class PartiallyRelevant:
    rank: Any


def get_rank_variable_relevancy(einsum: Einsum, tensor: TensorName):
    relevancy = {}
    projection = einsum.tensor_accesses[tensor].projection
    for rank_variable in einsum.rank_variables:
        relevancy[rank_variable] = Irrelevant()
        for rank_name, projection_str in projection.items():
            projection_expr = parse_expr(projection_str)
            is_simple = len(sympy.Add.make_args(projection_expr)) == 1
            is_relevant = (
                sympy.symbols(f"{rank_variable}") in projection_expr.free_symbols
            )

            if not is_relevant:
                continue

            if is_simple:
                relevancy[rank_variable] = Relevant(rank=rank_name)
            else:
                relevancy[rank_variable] = PartiallyRelevant(rank=rank_name)

            break
    return relevancy


def compute_dense_tile_occupancy(
    projection_expr: dict[str, sympy.Expr], rank_variable_shapes: dict
):
    result = 1
    for index_expr in projection_expr.values():
        subs = {
            s: rank_variable_shapes[s.name] - 1
            for s in index_expr.free_symbols
            if s.name in rank_variable_shapes
        }
        result = result * ((index_expr.xreplace(subs) if subs else index_expr) + 1)
    return result


def compute_rank_occupancy(projection_expr: sympy.Expr, rank_variable_shapes: dict):
    subs = {
        s: rank_variable_shapes[s.name] - 1
        for s in projection_expr.free_symbols
        if s.name in rank_variable_shapes
    }
    return (projection_expr.xreplace(subs) if subs else projection_expr) + 1


def get_stride_and_halo_of_einsum(
    einsum_name: str,
    workload: Workload,
    rank_variable_bounds: dict[RankVariable, int] | None = None,
) -> dict[TensorName, dict[tuple[Rank, RankVariable]], tuple[int, int]]:
    """
    Get stride and halo (initial delta) for an Einsum in workload.

    Returns dictionary mapping tensor to another dictionary mapping
    (rank, rank_var) to the stride and halo.
    """
    stride_and_halo = {}
    einsum = workload.einsums[einsum_name]
    if rank_variable_bounds is None:
        shape = get_rank_variable_bounds(workload, einsum_name)
    else:
        shape = rank_variable_bounds
    for tensor in einsum.tensor_names:
        stride_and_halo[tensor] = {}
        tensor_stride_and_halo = stride_and_halo[tensor]

        projection = get_projection_expr(einsum, tensor)
        tensor_accesses = einsum.tensor_accesses[tensor]
        for rank, rank_vars in tensor_accesses.rank2rank_variables.items():
            rank_projection = projection[rank]
            for rank_var in rank_vars:
                stride = rank_projection.coeff(rank_var)

                # Careful: in-place mutation of cons_shape
                original_shape = shape[rank_var]
                shape[rank_var] = 1
                halo = compute_rank_occupancy(rank_projection, shape) - 1
                shape[rank_var] = original_shape

                tensor_stride_and_halo[(rank, rank_var)] = (stride, halo)
    return stride_and_halo


def get_stride_and_halo(
    workload: Workload,
) -> dict[
    tuple[EinsumName, TensorName],
    dict[tuple[Rank, RankVariable], tuple[int, int]],
]:
    """
    Get stride and halo (initial delta) for Einsums in workload.

    Returns dictionary mapping (Einsum, tensor) to another dictionary mapping
    (rank, rank_var) to the stride and halo.
    """
    stride_and_halo = {}
    for einsum in workload.einsums:
        stride_and_halo_of_einsum = get_stride_and_halo_of_einsum(einsum.name, workload)
        for tensor, ranks2stride_and_halo in stride_and_halo_of_einsum.items():
            stride_and_halo[(einsum.name, tensor)] = ranks2stride_and_halo
    return stride_and_halo



class FlattenedArch:
    """
    A flattened arch is an architecture spec that has been
    flattened into a hierarchy for the purpose of mapping
    a particular Einsum.

    Several steps (may not be exhaustive) are applied when
    an arch is flattened:
    - A compute unit is selected, and a flattened arch
      is a path from the root to that compute unit
    - Expressions have been evaluated in the context of an
      Einsum.
    - Non-hierarchical arch nodes (e.g., ones inside the
      `nodes` key of an `Array` has been inserted into
      a hierarchy.)

    This class should only be relevant to the model and
    mapper. That is, the user should generally not define
    a flattened arch directly. So, unlike other classes in
    `frontend`, this one is intentionally *not* an
    `EvalableModel`.
    """

    def __init__(self, nodes: list["Leaf"]):
        self.nodes = nodes

    def __getitem__(self, idx: int | str | slice):
        if isinstance(idx, (int, slice)):
            return self.nodes[idx]
        elif isinstance(idx, str):
            for node in self.nodes:
                if node.name == idx:
                    return node
            raise KeyError(f"arch node with name {idx} not found")
        raise ValueError(f"idx should be int or str, but instead {type(idx)}")

    def __iter__(self):
        for node in self.nodes:
            yield node

    def index(self, name: str):
        for i, node in enumerate(self.nodes):
            if node.name == name:
                return i
        raise ValueError(f"no node found with name {name}")

    def is_above(self, name_a: str, name_b: str):
        """
        Returns True if node with name_a is above node with name_b.
        Raises ValueError if either is not found.
        """
        idx_a = self.index(name_a)
        idx_b = self.index(name_b)
        return idx_a < idx_b



import os
from pathlib import Path

# Use __file__ to get absolute paths relative to this file's location
_TESTS_DIR = Path(__file__).parent
_REPO_ROOT = _TESTS_DIR.parent

EXAMPLES_DIR = _REPO_ROOT / "examples"
CURRENT_DIR = _TESTS_DIR
NOTEBOOKS_DIR = _REPO_ROOT / "notebooks"



from math import ceil, log2
from hwcomponents import ComponentModel, action
from hwcomponents.scaling import linear, quadratic, reciprocal
from hwcomponents_library.library.aladdin import AladdinComparator, AladdinCounter
from hwcomponents_neurosim import FlipFlop


class ZeroComparator(ComponentModel):
    """
    Counts the number of zeros in a list of values. Includes a flag for each zero.

    Based on the zero gating logic in the paper: A Programmable Heterogeneous
    Microprocessor Based on Bit-Scalable In-Memory Computing, by Hongyang Jia, Hossein
    Valavi, Yinqi Tang, Jintao Zhang, and Naveen Verma, JSSC 2020
    10.1109/JSSC.2020.2987714

    Parameters
    ----------
    n_comparators: int
        The number of comparators to include.
    n_bits: int
        The number of bits of each comparator.
    tech_node: str
        The technology node of the comparators.
    voltage: float
        The voltage of the comparators.
    """

    priority = 0.5

    def __init__(
        self, n_comparators: int, n_bits: int, tech_node: str, voltage: float = 0.85
    ):
        self.n_comparators = n_comparators
        self.n_bits = n_bits

        # Scale up the comparator to handle all the comparators
        self.comparator = AladdinComparator(
            n_bits=n_bits,
            tech_node=tech_node,
        )
        self.comparator.energy_scale *= n_comparators
        self.comparator.area_scale *= n_comparators

        # Flip flops are used one bit at a time, so we only make one bit and scale the
        # energy and latency
        self.flip_flop = FlipFlop(
            n_bits=1,
            tech_node=tech_node,
        )
        self.flip_flop.energy_scale *= n_bits
        self.flip_flop.latency_scale *= n_bits

        # Zero counter is shared between all the comparators, so scale the energy and
        # latency to activate with each one
        self.zeros_counter = AladdinCounter(
            n_bits=ceil(log2(n_comparators)),
            tech_node=tech_node,
        )
        self.zeros_counter.energy_scale *= n_comparators
        self.zeros_counter.latency_scale *= n_comparators

        super().__init__(
            subcomponents=[
                self.comparator,
                self.flip_flop,
                self.zeros_counter,
            ],
        )

        for subcomponent in self.subcomponents:
            subcomponent.scale(
                "voltage",
                voltage,
                0.85,
                area_scale_function=linear,
                energy_scale_function=quadratic,
                latency_scale_function=reciprocal,
                leak_power_scale_function=linear,
            )
            subcomponent.leak_power_scale *= 0.02  # Low-leakage technology

    @action
    def read(self) -> tuple[float, float]:
        self.comparator.read()
        self.flip_flop.read()
        self.zeros_counter.read()



from dataclasses import dataclass
from typing import Any


@dataclass(frozen=True)
class ActionKey:
    level: str
    action: str


@dataclass(frozen=True)
class VerboseActionKey(ActionKey):
    tensor: str | None
    einsum: str


@dataclass
class ActionCount:
    total: Any
    max_per_unit: Any

    @staticmethod
    def default():
        return ActionCount(0, 0)



from __future__ import annotations

from accelforge.frontend import arch
from accelforge.frontend import config
from accelforge.frontend import mapping
from accelforge.frontend import renames
from accelforge.frontend import spec
from accelforge.frontend import variables
from accelforge.frontend import workload
from accelforge.frontend.spec import Spec, Spec
from accelforge.mapper.FFM import Metrics
from accelforge.util import set_n_parallel_jobs
from accelforge.util import LiteralString
import accelforge.mapper as mapper
from accelforge.examples import examples
from accelforge.util.exceptions import EvaluationError

from accelforge.frontend.variables import Variables
from accelforge.frontend.arch import Arch
from accelforge.frontend.config import Config
from accelforge.frontend.mapping import Mapping
from accelforge.frontend.renames import Renames
from accelforge.frontend.spec import Spec
from accelforge.frontend.workload import Workload

__all__ = [
    # Submodules
    "arch",
    "config",
    "mapping",
    "mapper",
    "renames",
    "spec",
    "variables",
    "workload",
    # Main classes
    "Arch",
    "Config",
    "Mapping",
    "Metrics",
    "Renames",
    "Spec",
    "Variables",
    "Workload",
    # Utilities
    "LiteralString",
    "set_n_parallel_jobs",
    "examples",
    "EvaluationError",
]



import math
from typing import List


def rescale_sum_to_1(array: List[float], do_not_change_index: int = -1) -> List[float]:
    """Rescales all list elements such that the sum is 1."""
    sum_array = sum([a for i, a in enumerate(array) if i != do_not_change_index])
    target_sum = 1 - array[do_not_change_index] if do_not_change_index >= 0 else 1
    scaleby = target_sum / sum_array
    return [a * scaleby if i != do_not_change_index else a for i, a in enumerate(array)]


def set_element_rescale_sum_to_1(array: List[float], index: int, value: float):
    """
    Sets an element of a list, then rescales all list elements such that the sum is 1.
    """
    array[index] = value
    return rescale_sum_to_1(array, index)


def value2bits(value: int, resolution: int) -> List[int]:
    """Converts a value to a list of bits."""
    return [int(i) for i in bin(value)[2:].zfill(resolution)]


def bit_distribution_2_hist(
    bit_distribution: List[float], zero_prob: float = None
) -> List[float]:
    """Converts a bit distribution to a value distribution."""
    hist = [1] * 2 ** len(bit_distribution)
    for value in range(2 ** len(bit_distribution)):
        bits = value2bits(value, len(bit_distribution))
        for i, prob in enumerate(bit_distribution):
            hist[value] *= prob if bits[i] else 1 - prob

    if zero_prob is not None:
        set_element_rescale_sum_to_1(hist, 0, zero_prob)
    return rescale_sum_to_1(hist)


def hist_2_bit_distribution(hist: List[float]) -> List[float]:
    """Converts a value distribution to a bit distribution."""
    sum_hist = sum(hist)
    hist = [i / sum_hist for i in hist]

    bit_distribution = [0] * math.ceil(math.log(len(hist), 2))
    for value in range(len(hist)):
        for i, bit in enumerate(value2bits(value, len(bit_distribution))):
            bit_distribution[i] += hist[value] * bit

    return bit_distribution



"""
Tests some of the key custom isl functions.
"""

import unittest

import islpy as isl

from accelforge.model._looptree.reuse.isl.isl_functions import *


class BasicIslFunctionTests(unittest.TestCase):
    """
    Prima facie checks of `isl_functions` to ensure basic functionality.
    """

    def test_project_dim_after(self):
        pass



from typing import (
    Any,
)

from accelforge.util._basetypes import (
    EvalExtras,
    _PostCall,
)

from accelforge.frontend.arch.structure import Leaf, Branch, Hierarchical
from accelforge.frontend.arch.components import Component
from accelforge.frontend.arch.spatialable import Spatialable, Spatial


class Arch(Hierarchical):
    """
    Top-level architecture specification.

    All attributes in the architecture can refrence variables in the spec-level
    `variables` field as well as symbols from the individual Einsum being processed.
    """

    variables: EvalExtras = EvalExtras()
    """
    Like the spec-level `variables` field, this field is evaluated first and its contents
    can be referenced elsewhere in the architecture. Unlike the spec-level `variables`
    field, this, like ther rest of the architecture, is evaluated per-Einsum and can
    reference Einsum-specific symbols.
    """

    extra_attributes_for_all_component_models: EvalExtras = EvalExtras()
    """
    Extra attributes to pass to all component models. This can be used to pass global
    attributes, such as technology node or clock period, to every component model.
    """

    @property
    def total_area(self) -> float:
        """
        Returns the total area of the architecture in m^2.

        Returns
        -------
        float
            The total area of the architecture in m^2.
        """
        return sum(self.per_component_total_area.values())

    @property
    def total_leak_power(self) -> float:
        """
        Returns the total leak power of the architecture in W.

        Returns
        -------
        float
            The total leak power of the architecture in W.
        """
        return sum(self.per_component_total_leak_power.values())

    @property
    def per_component_total_area(self) -> dict[str, float]:
        """
        Returns the total area used by each component in the architecture in m^2.

        Returns
        -------
        dict[str, float]
            A dictionary of component names to their total area in m^2.
        """
        area = {
            node.name: node.total_area for node in self.get_nodes_of_type(Component)
        }
        for k, v in area.items():
            if v is None:
                raise ValueError(
                    f"Area of {k} is not set. Please call the Spec's "
                    "`calculate_component_area_energy_latency_leak` method before accessing this "
                    "property."
                )
        return area

    @property
    def per_component_total_leak_power(self) -> dict[str, float]:
        """
        Returns the total leak power of each component in the architecture in W.

        Returns
        -------
        dict[str, float]
            A dictionary of component names to their total leak power in W.
        """
        leak_power = {
            node.name: node.total_leak_power
            for node in self.get_nodes_of_type(Component)
        }
        for k, v in leak_power.items():
            if v is None:
                raise ValueError(
                    f"Leak power of {k} is not set. Please call the Spec's "
                    "`calculate_component_area_energy_latency_leak` method before accessing this "
                    "property."
                )
        return leak_power

    def _eval_expressions(self, symbol_table: dict[str, Any], *args, **kwargs):
        outer_st = symbol_table

        class PostCallArch(_PostCall):
            def __call__(self, field, value, evaluated, symbol_table):
                if field == "variables":
                    # We're going to override the spec-level "variables", so make sure
                    # we copy over all the symbols from the spec-level "variables".
                    evaluated_dump = evaluated.shallow_model_dump()
                    for k, v in symbol_table.get("variables", {}).items():
                        if k not in evaluated_dump:
                            evaluated_dump[k] = v
                    symbol_table.update(evaluated_dump)
                    symbol_table["variables"] = evaluated_dump
                if field == "extra_attributes_for_all_component_models":
                    evaluated_dump = evaluated.shallow_model_dump()
                    symbol_table["arch_extra_attributes_for_all_component_models"] = (
                        evaluated_dump
                    )
                return evaluated

        cur_st = dict(symbol_table)

        for node in self.get_nodes_of_type(Leaf):
            cur_st[node.name] = node

        evaluated, _ = super()._eval_expressions(
            cur_st,
            *args,
            **kwargs,
            post_calls=(PostCallArch(),),
            order=(
                "variables",
                "extra_attributes_for_all_component_models",
            ),
        )
        return evaluated, symbol_table

    def __getitem__(self, name: str) -> Leaf:
        attr_error = None
        find_error = None
        try:
            return super().__getitem__(name)
        except KeyError as e1:
            attr_error = e1
        try:
            return self.find(name)
        except ValueError as e2:
            find_error = e2

        e1_str = "\n\t" + str(attr_error).replace("\n", "\n\t")
        e2_str = "\n\t" + str(find_error).replace("\n", "\n\t")

        errstr = (
            f"Could not access {name} in {self}, either as an attribute or as the name "
            f"of a leaf node. Error for accessing as an attribute:{e1_str}\n"
            f"Error for accessing as the name of a leaf node:{e2_str}"
        )

        raise KeyError(errstr)

    def find_spatial(
        self, name: str, return_spatialable: bool = False
    ) -> Spatial | tuple[Spatialable, Spatial]:
        """
        Find a spatial dimension by name. Raises an error if zero or more than one
        spatial dimension has the given name.

        Parameters
        ----------
        name: str
            The name of the spatial dimension to find.
        return_spatialable: bool
            Whether to return the node that the spatial dimension is associated with.

        Returns
        -------
        Spatial
            The spatial dimension with the given name if `return_spatialable` is False.
        tuple[Spatialable, Spatial]
            The node and spatial dimension with the given name if `return_spatialable`
            is True.

        Raises
        ------
        ValueError
            If no spatial dimension with the given name exists, or if more than one
            spatial dimension with the given name exists.
        """

        matches = []
        for node in self.get_nodes_of_type(Spatialable):
            for s in node.spatial:
                if s.name == name:
                    matches.append((node, s))

        if len(matches) == 0:
            raise ValueError(
                f"No spatial dimension with name '{name}' found in the architecture"
            )
        if len(matches) > 1:
            node_names = [getattr(n, "name", type(n).__name__) for n, _ in matches]
            raise ValueError(
                f"Multiple spatials with name '{name}' found in the architecture "
                f"(in nodes: {node_names}). Expected exactly one."
            )
        if return_spatialable:
            return matches[0]
        return matches[0][1]

    def model_post_init(self, __context__=None) -> None:
        # Make sure all leaf names are unique
        leaves = {}
        for l in self.get_nodes_of_type(Leaf):
            n = l.name
            leaves.setdefault(n, l)
            assert l is leaves[n], f"Duplicate name {n} found in architecture"



from accelforge.mapper.FFM.main import (
    map_workload_to_arch,
    make_pmappings,
    join_pmappings,
    MultiEinsumPmappings,
    Mappings,
)
from accelforge.frontend.mapper.metrics import Metrics
from accelforge.mapper.FFM._join_pmappings.pmapping_group import PmappingGroup

__all__ = [
    "map_workload_to_arch",
    "make_pmappings",
    "join_pmappings",
    "MultiEinsumPmappings",
    "Mappings",
    "Metrics",
    "PmappingGroup",
]



from typing import Iterable
from numbers import Number

import matplotlib.pyplot as plt

from accelforge.mapper.FFM import Mappings


def plot_roofline(
    bandwidth: Number,
    computational_throughput: Number,
    min_computational_intensity: Number = 0,
    max_computational_intensity: Number = None,
):
    """
    Plot a roofline model.

    Parameters
    ----------
    bandwidth:
        The memory bandwidth to use when generating the roofline.
    computational_throughput:
        The peak computational throughput to use when generating the roofline.
    min_computational_intensity:
        The minimum computational intensity to include in the x-axis.
    max_computational_intensity:
        The maximum computational intensity to include in the x-axis.
    """
    fig, ax = plt.subplots()

    roofline_transition = _roofline_transition(bandwidth, computational_throughput)
    if max_computational_intensity is None:
        max_computational_intensity = 2 * roofline_transition

    ax.plot(
        [min_computational_intensity, roofline_transition],
        [min_computational_intensity * bandwidth, computational_throughput],
        color="black",
    )
    ax.plot(
        [roofline_transition, max_computational_intensity],
        [computational_throughput, computational_throughput],
        color="black",
    )

    ax.spines.right.set_visible(False)
    ax.spines.top.set_visible(False)

    return fig, ax


def _roofline_transition(bandwidth, computation_throughput):
    return computation_throughput / bandwidth
